import paddle
import paddle.nn.functional as F

from iann.inference.transforms import (
    AddHorizontalFlip,
    SigmoidForPred,
    LimitLongestSide,
)


class BasePredictor(object):
    def __init__(
        self,
        net,
        net_clicks_limit=None,
        with_flip=False,
        zoom_in=None,
        max_size=None,
        **kwargs
    ):
        self.net = net
        self.with_flip = with_flip
        self.net_clicks_limit = net_clicks_limit
        self.original_image = None
        self.zoom_in = zoom_in

        self.transforms = [zoom_in] if zoom_in is not None else []
        # self.transforms =  []
        if max_size is not None:
            self.transforms.append(LimitLongestSide(max_size=max_size))
        self.transforms.append(SigmoidForPred())
        if with_flip:
            self.transforms.append(AddHorizontalFlip())

    def set_input_image(self, image_nd):
        for transform in self.transforms:
            transform.reset()
        # print('imagend', image_nd.shape)
        self.original_image = paddle.to_tensor(image_nd).astype("float32")

        if len(self.original_image.shape) == 3:
            self.original_image = self.original_image.unsqueeze(0)
        # for x in self.transforms:
        #     print("&&&&&&&&", x.__class__.__name__)

    def get_prediction(self, clicker, mask=None):
        clicks_list = clicker.get_clicks()

        image_nd, clicks_lists, is_image_changed, mask = self.apply_transforms(
            self.original_image, [clicks_list], mask
        )
        pred_logits = self._get_prediction(
            image_nd=image_nd,
            clicks_lists=clicks_lists,
            is_image_changed=is_image_changed,
            # BUG: 这里下面的 _get_prediction 不识这个mask
            mask=mask,
        )
        prediction = F.interpolate(
            pred_logits, mode="bilinear", align_corners=True, size=image_nd.shape[2:]
        )

        for t in reversed(self.transforms):
            prediction = t.inv_transform(prediction)

        if self.zoom_in is not None and self.zoom_in.check_possible_recalculation():
            return self.get_prediction(clicker)
        # print('prediction', prediction.numpy()[0, 0])
        return prediction.numpy()[0, 0]

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed, mask):
        points_nd = self.get_points_nd(clicks_lists)
        return self.net(image_nd, points_nd)["instances"]

    def _get_transform_states(self):
        return [x.get_state() for x in self.transforms]

    def _set_transform_states(self, states):
        assert len(states) == len(self.transforms)
        for state, transform in zip(states, self.transforms):
            transform.set_state(state)

    def apply_transforms(self, image_nd, clicks_lists, mask):
        is_image_changed = False
        for t in self.transforms:
            image_nd, clicks_lists, mask = t.transform(image_nd, clicks_lists, mask)
            is_image_changed |= t.image_changed

        return image_nd, clicks_lists, is_image_changed, mask

    def get_points_nd(self, clicks_lists):
        total_clicks = []
        num_pos_clicks = [
            sum(x.is_positive for x in clicks_list) for clicks_list in clicks_lists
        ]
        num_neg_clicks = [
            len(clicks_list) - num_pos
            for clicks_list, num_pos in zip(clicks_lists, num_pos_clicks)
        ]
        num_max_points = max(num_pos_clicks + num_neg_clicks)
        if self.net_clicks_limit is not None:
            num_max_points = min(self.net_clicks_limit, num_max_points)
        num_max_points = max(1, num_max_points)

        for clicks_list in clicks_lists:
            clicks_list = clicks_list[: self.net_clicks_limit]
            pos_clicks = [click.coords for click in clicks_list if click.is_positive]
            pos_clicks = pos_clicks + (num_max_points - len(pos_clicks)) * [(-1, -1)]

            neg_clicks = [
                click.coords for click in clicks_list if not click.is_positive
            ]
            neg_clicks = neg_clicks + (num_max_points - len(neg_clicks)) * [(-1, -1)]
            total_clicks.append(pos_clicks + neg_clicks)

        return paddle.to_tensor(total_clicks)

    def get_states(self):
        return {"transform_states": self._get_transform_states()}

    def set_states(self, states):
        self._set_transform_states(states["transform_states"])
